'by Charles Dietz, released 09-24-03
'rev 09-28-03: added ABS(dataViewFocus) at end of WM_PAINT
'
'This include file defines a custom control to view and edit
'an array of data with scroll bars for large arrays
'
'CALL DataView(hParent, id, a(), x, y, w, h, nCols, nFont)
'  hParent = handle of dialog containing dataView control
'  id      = dataView control id
'  a()     = string values to display as a(row, col)
'  x, y    = are position coordinates for dataView control in dialog units
'  w, h    = are width and height of dataView control in dialog units
'  style   = [%DV_COLHDRS|%DV_ROWHDRS|%DV_VIEWONLY] + ffcc
'            ff, font size (8 if zero)
'            cc, number of columns shown (best fit if zero)
'            use %DV_COLHDRS to include column headers
'            use %DV_ROWHDRS to include row headers
'            use %DV_VIEWONLY to view without editing
'
'  Column headers are stored in a(0, col)
'  Row headers are stored in a(row, 0)
'
'  Navigate controls with mouse or keyboard
'       ... vertically with Up/Dn arrows keys
'       ... horizontally with PageUp/PageDn keys
'       ... home key to reset upper left corner
'       ... tab to set focus on next control
'---------------------------------------------------------------------------
%DV_COLHDRS = 4096
%DV_ROWHDRS = 8192
%DV_VIEWONLY = 16384

TYPE dataViewStore                     '96 bytes total
   hDlg AS LONG                        'control handle
   id AS LONG                          'control id
   hEdt AS LONG                        'editbox handle
   startX0 AS LONG                     'x position of 1st element
   startY0 AS LONG                     'y position of 1st element
   colWi AS LONG                       'column width
   rowHt AS LONG                       'row height
   arrPtr AS DWORD                     'pointer to input array
   dimI AS LONG                        'array row dim
   dimJ AS LONG                        'array col dim
   ncols AS LONG                       'number of columns/headers
   nFont AS LONG                       'font size for control
   memdc AS LONG                       'memory dc for control
   hBit AS LONG                        'bitmap handle for control
   hFont AS LONG                       'font handle for control
   wi AS LONG                          'width of dataView in pixels
   ht AS LONG                          'height of dataView in pixels
   viewII AS LONG                      'viewed rows excluding col header
   viewJJ AS LONG                      'viewed cols excluding row header
   disp AS LONG                        'control has been displayed
   row AS LONG                         'row position for editbox
   col AS LONG                         'col position for editbox
   Vpos AS LONG                        'vertical scroll position
   Hpos AS LONG                        'horizontal scroll positon
END TYPE

GLOBAL dataViewNum AS LONG             'number of active dataView controls
GLOBAL dataViewIndex AS LONG           'index for dataViewSto()
GLOBAL dataViewFocus AS LONG           'current control focus
GLOBAL dataViewEd AS STRING            'user edited array value
GLOBAL dataViewSto() AS dataViewStore  'dataView storage
GLOBAL dataViewStr() AS STRING         'dataView string array, overlay a()

DECLARE FUNCTION dataViewProc(BYVAL LONG, BYVAL LONG, BYVAL LONG, BYVAL LONG) AS LONG
DECLARE FUNCTION dataViewGetDlg(LONG) AS LONG
DECLARE FUNCTION MakeFont(LONG, STRING, BYVAL LONG, OPT BYVAL STRING) AS LONG
DECLARE FUNCTION dataViewSubClassEd(BYVAL LONG, BYVAL LONG, BYVAL LONG, BYVAL LONG) AS LONG
DECLARE FUNCTION dataViewRegister() AS LONG
DECLARE SUB DataViewAdd(LONG, LONG, a() AS STRING, LONG, LONG, LONG, LONG, OPT BYVAL LONG)
DECLARE SUB txtOut(LONG, BYVAL LONG, BYVAL LONG, STRING)
DECLARE SUB getTextSize(LONG, LONG, LONG, LONG, STRING)
DECLARE SUB dataViewInit(LONG, SCROLLINFO, SCROLLINFO)
DECLARE SUB dataViewShow(LONG, BYVAL LONG)
DECLARE SUB dataViewEdit(LONG)

SUB DataViewAdd(hDlg AS LONG, id AS LONG, a() AS STRING, x AS LONG, y AS LONG, _
       w AS LONG, h AS LONG, OPT BYVAL nopt AS LONG)
   LOCAL style AS LONG, nhdrs AS LONG, ncols AS LONG, nFont AS LONG
   LOCAL m AS LONG, n AS LONG, i AS LONG, j AS LONG
   LOCAL ww AS LONG, hh AS LONG
   FOR i = 1 TO dataViewNum 'check for prior id running
      IF dataViewSto(i).id = id THEN EXIT SUB
   NEXT i
   dataViewRegister
   INCR dataViewNum
   REDIM PRESERVE dataViewSto(dataViewNum) AS dataViewStore
   m = UBOUND(a, 1): n = UBOUND(a, 2)
   nFont = (nopt MOD 4096) \ 100 'font size
   IF nFont = 0 THEN nFont = 10
   dataViewSto(dataViewNum).nFont = nFont
   nhdrs = IIF(nopt >= 16384, nopt - 16384, nopt) \ 4096
   ncols = (nopt MOD 4096) MOD 100
   IF ncols = 0 THEN ncols = n
   i = nhdrs * 100 + ncols
   IF nopt >= 16384 THEN i = -i
   dataViewSto(dataViewNum).ncols = i
   dataViewSto(dataViewNum).id = id
   dataViewSto(dataViewNum).colWi = 0
   dataViewSto(dataViewNum).rowHt = 0
   dataViewSto(dataViewNum).arrPtr = VARPTR(a(0, 0))
   dataViewSto(dataViewNum).dimI = m
   dataViewSto(dataViewNum).dimJ = n
   'add the dataView custom control -------------------------------------
   style = %WS_CHILD OR %WS_VISIBLE OR %WS_TABSTOP OR %WS_BORDER
   getTextSize hDlg, i, j, nFont, "dataView"
   DIALOG UNITS hDlg, w, h TO PIXELS ww, hh
   i = hh \ j: IF nhdrs = 1 OR nhdrs = 3 THEN i = i - 2
   IF i < m THEN style = style OR %WS_VSCROLL
   IF ncols < n THEN style = style OR %WS_HSCROLL
   CONTROL ADD "dataViewCtrl", hDlg, id, "", x, y, w, h, style
END SUB

FUNCTION dataViewRegister() AS LONG
  LOCAL wc AS WNDCLASS, szClassName AS ASCIIZ * 13
  'Register custom control window class.
  szClassName      = "dataViewCtrl"
  wc.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_PARENTDC OR %CS_GLOBALCLASS
  wc.lpfnWndProc   = CODEPTR(dataViewProc)
  wc.cbClsExtra    = 0
  wc.cbWndExtra    = 0
  wc.hInstance     = GetModuleHandle(BYVAL %NULL)
  wc.hIcon         = %NULL
  wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_Arrow)
  wc.hbrBackground = %COLOR_WINDOW + 1
  wc.lpszMenuName  = %NULL
  wc.lpszClassName = VARPTR(szClassName)
  FUNCTION = RegisterClass(wc)
END FUNCTION

FUNCTION dataViewProc(BYVAL hWnd AS LONG, BYVAL wMsg AS LONG, _
             BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
   LOCAL ps AS PAINTSTRUCT, style AS LONG, m AS LONG, n AS LONG
   LOCAL x AS LONG, y AS LONG, xx AS LONG, yy AS LONG
   LOCAL memdc AS LONG, hdc AS LONG, hBit AS LONG, hBrush AS LONG
   LOCAL hFont AS LONG, hOldFont AS LONG, lpCS AS CREATESTRUCT PTR
   LOCAL w AS LONG, h AS LONG, x0 AS LONG, y0 AS LONG
   LOCAL hEdt AS LONG, orig AS LONG, szOrigWndProc AS ASCIIZ*12
   LOCAL vPos AS LONG, hPos AS LONG, hOldBit AS LONG
   STATIC siV AS SCROLLINFO, siH AS SCROLLINFO
   SELECT CASE wMsg
      CASE %WM_CREATE
         dataViewSto(dataViewNum).hDlg = hWnd
         dataViewInit hWnd, siV, siH
         w = dataViewSto(dataViewNum).wi
         h = dataViewSto(dataViewNum).ht
         hdc = GetDC(hWnd)
         memdc = CreateCompatibleDC(hdc)
         hBit = CreateCompatibleBitmap(hdc, w, h)
         SelectObject memdc, hBit
         hBrush = GetStockObject(%WHITE_BRUSH)
         SelectObject memdc, hBrush
         PatBlt memdc, 0, 0, w, h, %PATCOPY
         dataViewSto(dataViewNum).hbit = hBit
         ReleaseDC hWnd, hdc
         dataViewSto(dataViewNum).memdc = memdc
         dataViewFocus = -1 'flag to begin after all controls are shown
         'set up editbox for dataView controls -----------------------------
         w = dataViewSto(dataViewNum).colWi - 4
         h = dataViewSto(dataViewNum).rowHt + 2
         style = %WS_CHILD OR %WS_VISIBLE OR %WS_BORDER OR %ES_LEFT
         lpCS = lParam
         hEdt = CreateWindow("edit", BYVAL %NULL, style, -99, -99, w, h,_
                             hWnd, n, @lpCS.hInstance, BYVAL %NULL)
         dataViewSto(dataViewNum).hEdt = hEdt
         'specify font for editbox -----------------------------------------
         hdc = GetDC(hWnd)
         n = dataViewSto(dataViewNum).nFont
         hFont = makeFont(hdc, "MS Sans Serif", n)
         SelectObject hdc, hFont
         SendMessage hEdt, %WM_SETFONT, hFont, 0
         dataViewSto(dataViewNum).hFont = hFont
         ReleaseDC hWnd, hdc
         'subclass the editbox ---------------------------------------------
         orig = SetWindowLong(hEdt, %GWL_WNDPROC, CODEPTR(dataViewSubClassEd))
         szOrigWndProc = "OrigWndProc"
         SetProp hEdt, szOrigWndProc, orig
      CASE %WM_NCPAINT 'restore controls
         n = dataViewGetDlg(hWnd): dataViewSto(n).disp = 0
         IF n = dataViewFocus THEN SetFocus hWnd
         UpdateWindow hWnd
      CASE %WM_KILLFOCUS, %WM_SETFOCUS
         dataViewSto(dataViewGetDlg(hWnd)).disp = 0
         InvalidateRect hWnd, BYVAL %NULL, 0
      CASE %WM_PAINT
         n = dataViewGetDlg(hWnd): m = dataViewIndex
         IF n = 0 THEN EXIT SELECT
         IF n <> m THEN dataViewInit(hWnd, siV, siH)
         IF dataViewSto(n).disp THEN EXIT SELECT
         hdc = BeginPaint(hWnd, ps)
         w = dataViewSto(n).wi: h = dataViewSto(n).ht
         memdc = dataViewSto(n).memdc
         dataViewShow hWnd, memdc
         BitBlt hdc, 0, 0, w, h, memdc, 0, 0, %SRCCOPY
         EndPaint hWnd, ps
         IF dataViewFocus < 0 AND dataViewIndex = dataViewNum THEN
            'begin... all controls shown now
            n = dataViewSto(ABS(dataViewFocus)).hDlg
            dataViewInit(n, siV, siH)
         END IF
         dataViewFocus = ABS(dataViewFocus)
      CASE %WM_KEYUP
         SELECT CASE wParam
            CASE %VK_UP
               SendMessage hWnd, %WM_VSCROLL, %SB_LINEUP, 0
            CASE %VK_DOWN
               SendMessage hWnd, %WM_VSCROLL, %SB_LINEDOWN, 0
         END SELECT
      CASE %WM_VSCROLL
         n = dataViewGetDlg(hWnd)
         dataViewSto(n).disp = 0
         vPos = dataViewSto(n).vPos
         SELECT CASE LOWRD(wParam)
            CASE %SB_TOP
               vPos = 1
            CASE %SB_LINEDOWN
               INCR vPos
               IF vPos > siV.nMax THEN vPos = siV.nMax: EXIT FUNCTION
            CASE %SB_LINEUP
               DECR vPos
               IF vPos < siV.nMin THEN vPos = siV.nMin: EXIT FUNCTION
            CASE %SB_THUMBTRACK
               vPos = HIWRD(wParam)
               IF vPos < siV.nMin THEN vPos = siV.nMin: EXIT FUNCTION
               IF vPos > siV.nMax THEN vPos = siV.nMax: EXIT FUNCTION
            CASE %SB_PAGEDOWN
               vPos = vPos + (siV.nMax - siV.nMin)\10 + 1
               IF vPos > siV.nMax THEN vPos = siV.nMax: EXIT FUNCTION
            CASE %SB_PAGEUP
               vPos = vPos - (siV.nMax - siV.nMin)\10 - 1
               IF vPos < siV.nMin THEN vPos = siV.nMin: EXIT FUNCTION
         END SELECT
         m = dataViewFocus
         IF n <> dataViewFocus THEN
            IF m THEN SetFocus dataViewSto(m).hDlg
         ELSE
            siV.fMask = %SIF_POS: siV.nPos = vPos
            SetScrollInfo BYVAL hWnd, BYVAL %SB_VERT, siV, 1
            InvalidateRect hWnd, BYVAL %NULL, 1
         END IF
         SetFocus hWnd
         dataViewSto(n).vPos = vPos
      CASE %WM_HSCROLL
         n = dataViewGetDlg(hWnd)
         dataViewSto(n).disp = 0
         hPos = dataViewSto(n).hPos
         SELECT CASE LOWRD(wParam)
            CASE %SB_TOP
               hPos = 1
            CASE %SB_LINERIGHT
               INCR hPos
               IF hPos > siH.nMax THEN hPos = siH.nMax: EXIT FUNCTION
            CASE %SB_LINELEFT
               DECR hPos
               IF hPos < siH.nMin THEN hPos = siH.nMin: EXIT FUNCTION
            CASE %SB_THUMBTRACK
               hPos = HIWRD(wParam)
               IF hPos < siH.nMin THEN hPos = siH.nMin: EXIT FUNCTION
               IF hPos > siH.nMax THEN hPos = siH.nMax: EXIT FUNCTION
            CASE %SB_PAGERIGHT
               hPos = hPos + (siH.nMax - siH.nMin)\10 + 1
               IF hPos > siH.nMax THEN hPos = siH.nMax: EXIT FUNCTION
            CASE %SB_PAGELEFT
               hPos = hPos - (siH.nMax - siH.nMin)\10 - 1
               IF hPos < siH.nMin THEN hPos = siH.nMin: EXIT FUNCTION
         END SELECT
         m = dataViewFocus
         IF n <> dataViewFocus THEN
            IF m THEN SetFocus dataViewSto(m).hDlg
         ELSE
            siH.fMask = %SIF_POS: siH.nPos = hPos
            SetScrollInfo BYVAL hWnd, BYVAL %SB_HORZ, siH, 1
            InvalidateRect hWnd, BYVAL %NULL, 1
         END IF
         SetFocus hWnd
         dataViewSto(n).hPos = hPos
      CASE %WM_LBUTTONUP
         n = dataViewGetDlg(hWnd): m = dataViewFocus
         IF n = dataViewFocus THEN
            IF dataViewSto(n).nCols < 0 THEN EXIT FUNCTION 'just viewing
            x = LOWRD(lParam): y = HIWRD(lParam)
            xx = dataViewSto(n).colWi
            yy = dataViewSto(n).rowHt
            x0 = dataViewSto(n).startX0
            y0 = dataViewSto(n).startY0
            dataViewSto(n).col = (x - x0) \ xx + dataViewSto(n).hPos
            dataViewSto(n).row = (y - y0) \ yy + dataViewSto(n).vPos
         ELSEIF m THEN
            dataViewSto(m).disp = 0
            InvalidateRect dataViewSto(m).hDlg, BYVAL %NULL, 1
            UpdateWindow dataViewSto(m).hDlg
         END IF
         SetFocus hWnd
      CASE %WM_DESTROY
         n = dataViewGetDlg(hWnd)
         DeleteDC dataViewSto(n).memdc
         DeleteObject dataViewSto(n).hBit
         DeleteObject dataViewSto(n).hFont
         ARRAY DELETE dataViewSto(n): DECR dataViewNum
   END SELECT
   'default processing for other messages.
   FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)
END FUNCTION

SUB dataViewInit(hdv AS LONG, siV AS SCROLLINFO, siH AS SCROLLINFO)
   'setup and show newly selected dataView control -------------------------
   LOCAL i AS LONG, k AS LONG, m AS LONG, n AS LONG, nOpt AS LONG
   LOCAL w AS LONG, h AS LONG, x AS LONG, y AS LONG, r AS RECT
   LOCAL txtHt AS LONG, startX0 AS LONG, startY0 AS LONG, nFont AS LONG
   k = dataViewGetDlg(hdv)
   m = dataViewSto(k).dimI
   n = dataViewSto(k).dimJ
   REDIM dataViewStr(m, n) AT dataViewSto(k).arrPtr
   IF dataViewSto(k).colWi = 0 THEN 'new control
      GetClientRect hdv, r
      w = r.nRight - r.nLeft: h = r.nBottom - r.nTop
      dataViewSto(k).wi = w: dataViewSto(k).Ht = h
      nFont = dataViewSto(k).nFont
      nopt = ABS(dataViewSto(k).ncols) \ 100
      'get maximum width for row headers and beginning coords
      startX0 = 8
      IF nopt > 1 THEN
         startX0 = w / ((ABS(dataViewSto(k).ncols) MOD 100) + 1)
         FOR i = 0 TO m
            getTextSize hdv, x, y, nFont, dataViewStr(i, 0)
            startX0 = MAX(x + 8, startX0)
         NEXT i
      END IF
      getTextSize hdv, x, txtHt, nFont, "dataView"
      dataViewSto(k).startX0 = startX0
      IF nOpt = 1 OR nopt = 3 THEN startY0 = 2 * txtHt
      startY0 = startY0 + 4
      dataViewSto(k).startY0 = startY0
      i = (h - startY0) \ txtHt
      dataViewSto(k).viewII = MIN(i, m)
      dataViewSto(k).rowHt = txtHt
      i = ABS(dataViewSto(k).ncols) MOD 100
      dataViewSto(k).viewJJ = i
      dataViewSto(k).colWi = (w - startX0) / i
      dataViewSto(k).row = 1: dataViewSto(k).col = 1
      dataViewSto(k).vPos = 1: dataViewSto(k).hPos = 1
      EXIT SUB
   END IF
   dataViewIndex = k
   'setup vertical scrolling... 1st, calculate lines in window -------------
   i = dataViewSto(k).viewII
   siV.cbSize = SIZEOF(siV): siV.fMask = %SIF_ALL
   siV.nMin = 1: siV.nMax = m - i + 1
   siV.nPos = dataViewSto(k).vPos
   SetScrollInfo BYVAL hdv, BYVAL %SB_VERT, siV, 1
   'setup horizontal scrolling... 1st, calculate columns in window ---------
   i = dataViewSto(k).viewJJ
   siH.cbSize = SIZEOF(siH): siH.fMask = %SIF_ALL
   siH.nMin = 1: siH.nMax = n - i + 1
   siH.nPos = dataViewSto(k).hPos
   SetScrollInfo BYVAL hdv, BYVAL %SB_HORZ, siH, 1
END SUB

SUB dataViewShow(hdv AS LONG, BYVAL memdc AS LONG)
   LOCAL hFont AS LONG, hOldFont AS LONG
   LOCAL i AS LONG, j AS LONG, k AS LONG, m AS LONG, n AS LONG
   LOCAL x AS LONG, y AS LONG, w AS LONG, h AS LONG, nFont AS LONG
   LOCAL startX0 AS LONG, startY0 AS LONG, colWi AS LONG, rowHt AS LONG
   LOCAL nHdrs AS LONG, r AS RECT, viewII AS LONG, viewJJ AS LONG
   k = dataViewIndex: dataViewSto(k).disp = 1
   m = dataViewSto(k).dimI: n = dataViewSto(k).dimJ
   startX0 = dataViewSto(k).startX0: startY0 = dataViewSto(k).startY0
   colWi = dataViewSto(k).colWi: rowHt = dataViewSto(k).rowHt
   viewII = dataViewSto(k).viewII: viewJJ = dataViewSto(k).viewJJ
   nFont = dataViewSto(k).nFont
   hFont = makeFont(memdc, "MS Sans Serif", nFont, "u")
   hOldFont = SelectObject(memdc, hFont)
   nHdrs = ABS(dataViewSto(k).ncols) \ 100

   'clear window rect in pixels --------------------------------------------
   GetClientRect hdv, r
   w = r.nRight - r.nLeft: h = r.nBottom - r.nTop
   FillRect memdc, r, GetStockObject(%WHITE_BRUSH)
   IF hdv = GetFocus THEN
      'show double line border and set focus to edit box of control
      MoveToEx memdc, 1, 1, BYVAL %NULL
      LineTo memdc, 1, h
      MoveToEx memdc, 1, 1, BYVAL %NULL
      LineTo memdc, w, 1
      dataViewFocus = IIF(dataViewFocus < 0, -k, k)
      x = dataViewSto(k).hEdt
      SetFocus x
      SendMessage x, %EM_SETSEL, 0, -1
   END IF
   'show corner header if specified ----------------------------------------
   x = 8: y = 4
   IF nHdrs = 3 THEN txtOut memdc, x, y, dataViewStr(0, 0)

   'show column headers as underlined if specified -------------------------
   y = 4
   IF nHdrs = 1 OR nHdrs = 3 THEN
      x = startX0
      FOR j = dataViewSto(k).hPos TO n
         txtOut memdc, x, y, dataViewStr(0, j): x = x + colWi
         IF j = dataViewSto(k).hPos + viewJJ - 1 THEN EXIT FOR
      NEXT j
   END IF

   'show row headers if specified ------------------------------------------
   hFont = makeFont(memdc, "MS Sans Serif", nFont)
   DeleteObject selectObject(Memdc, hFont)
   IF nHdrs > 1 THEN
      x = 8: y = startY0
      FOR i = dataViewSto(k).vPos TO m
         txtOut memdc, x, y, dataViewStr(i, 0): y = y + rowHt
         IF i = dataViewSto(k).vPos + viewII - 1 THEN EXIT FOR
      NEXT i
   END IF

   'show array string elements ---------------------------------------------
   x = startX0: y = startY0
   FOR j = dataViewSto(k).hPos TO n
      FOR i = dataViewSto(k).vPos TO m
         txtOut memdc, x, y, dataViewStr(i, j): y = y + rowHt
         IF i = dataViewSto(k).vPos + viewII - 1 THEN EXIT FOR
      NEXT i
      x = x + colWi: y = startY0
      IF j = dataViewSto(k).hPos + viewJJ - 1 THEN EXIT FOR
   NEXT j
   DeleteObject selectObject(MemDc, hOldFont)
   i = dataViewSto(k).row: j = dataViewSto(k).col
   IF dataViewSto(k).ncols > 0 AND (i > 0 AND j > 0) THEN
      dataViewEdit dataViewSto(k).hEdt
   END IF
END SUB

SUB txtOut(hDC AS LONG, BYVAL x AS LONG, BYVAL y AS LONG, txt AS STRING)
   'x,y are the starting point coordinates in pixels
   TextOut hDC, x, y, BYVAL STRPTR(txt), LEN(txt)
END SUB

SUB getTextSize(hDlg AS LONG, x AS LONG, y AS LONG, f AS LONG, txt AS STRING)
   'return x,y as the txt dimensions in pixels
   LOCAL cSize AS SIZEL
   LOCAL hdc AS LONG, hFont AS LONG, hOldFont AS LONG
   hdc = GetDC(hDlg)
   hFont = makeFont(hdc, "MS Sans Serif", f)
   hOldFont = SelectObject(hdc, hFont)
   GetTextExtentExPoint BYVAL hdc, BYVAL STRPTR(txt), LEN(txt), 0, _
                        BYVAL %NULL, BYVAL %NULL, BYVAL VARPTR(cSize)
   x = cSize.cx: y = cSize.cy
   DeleteObject SelectObject(hdc, hOldFont)
   ReleaseDC hDlg, hdc
END SUB

FUNCTION dataViewGetDlg(hdv AS LONG) AS LONG
   'find the window corresponding to hdv
   LOCAL i AS LONG
   FOR i = 1 TO dataViewNum
      IF dataViewSto(i).hDlg = hdv THEN FUNCTION = i: EXIT FOR
   NEXT i
END FUNCTION

SUB dataViewEdit(hdv AS LONG)
   'this routine manages the control's edit box
   LOCAL i AS LONG, j AS LONG, ed AS LONG, m AS LONG, n AS LONG
   LOCAL x AS LONG, y AS LONG, x0 AS LONG, y0 AS LONG, r AS RECT
   LOCAL memdc AS LONG, hFont AS LONG, nFont AS LONG
   LOCAL viewII AS LONG, viewJJ AS LONG, startX0 AS LONG, startY0 AS LONG
   LOCAL colWi AS LONG, rowHt AS LONG, txt AS ASCIIZ * 256
   LOCAL k AS LONG, row AS LONG, col AS LONG
   'k = dataViewIndex
   k = dataViewGetDlg(GetParent(hdv))
   dataViewIndex = k
   memdc = dataViewSto(k).memdc
   viewII = dataViewSto(k).viewII
   viewJJ = dataViewSto(k).viewJJ
   startX0 = dataViewSto(k).startX0
   startY0 = dataViewSto(k).startY0
   colWi = dataViewSto(k).colWi
   rowHt = dataViewSto(k).rowHt
   IF LEN(dataViewEd) THEN ed = 1
   GetWindowRect GetParent(hdv), r: x0 = r.nLeft: y0 = r.nTop
   m = UBOUND(dataViewStr, 1): n = UBOUND(dataViewStr, 2)
   col = dataViewSto(k).col
   row = dataViewSto(k).row
   IF row < 0 OR col < 0 THEN
      ed = 0: row = ABS(row): col = ABS(col)
   END IF
   i = row - dataViewSto(k).vPos + 1
   j = col - dataViewSto(k).hPos + 1
   'next 4 statements keep textbox in viewing window
   IF i < 1 THEN i = 1: row = dataViewSto(k).vPos
   IF j < 1 THEN j = 1: col = dataViewSto(k).hPos
   IF i > viewII THEN i = viewII: row = dataViewSto(k).vPos + viewII - 1
   IF j > viewJJ THEN j = viewJJ: col = dataViewSto(k).hPos + viewJJ - 1
   x = startX0 + (j - 1) * colWi
   y = startY0 + (i - 1) * rowHt
   txt = dataViewStr(row, col)
   SetWindowPos hdv, 0, x, y, 0, 0, %SWP_NOSIZE OR %SWP_NOZORDER
   SetWindowText hdv, txt
   IF ed THEN  'edited text to display
      hFont = makeFont(memdc, "MS Sans Serif", nFont)
      SelectObject memdc, hFont
      DeleteObject hFont
      txtOut memdc, x0, y0, dataViewEd
   END IF
   dataViewEd = "": dataViewSto(k).row = row: dataViewSto(k).col = col
   IF dataViewFocus = dataViewIndex THEN SendMessage hdv, %EM_SETSEL, 0, -1
END SUB

FUNCTION dataViewSubClassEd(BYVAL hWnd&, BYVAL wMsg&, BYVAL wParam&, BYVAL lParam&) AS LONG
'capture keystrokes during edit
   LOCAL m AS LONG, n AS LONG, x AS LONG, x1 AS LONG, y AS LONG
   LOCAL szOrigWndProc AS ASCIIZ*12, hEdt AS LONG, hdv AS LONG
   LOCAL nFont AS LONG, colWi AS LONG, viewII AS LONG, viewJJ AS LONG
   LOCAL s AS STRING, txt AS ASCIIZ * 128, row AS LONG, col AS LONG
   szOrigWndProc = "OrigWndProc"
   SELECT CASE wMsg&
      CASE %WM_KILLFOCUS
         n = dataViewGetDlg(wParam&): m = dataViewFocus
         dataViewSto(n).disp = 0: dataViewSto(m).disp = 0
         IF n = 0 THEN
            InvalidateRect dataViewSto(m).hDlg, BYVAL %NULL, 0
            EXIT SELECT
         END IF
         IF n <> dataViewFocus THEN
            InvalidateRect dataViewSto(m).hDlg, BYVAL %NULL, 0
            UpdateWindow dataViewSto(m).hDlg
            SetFocus dataViewSto(n).hDlg
         END IF
      CASE %WM_LBUTTONUP
         SetFocus GetParent(hWnd&)
      CASE %WM_KEYUP
         SELECT CASE wParam&
            CASE %VK_RETURN, %VK_UP, %VK_DOWN, %VK_PGUP, %VK_PGDN, %VK_HOME
               n = dataViewGetDlg(GetParent(hWnd&))
               dataViewSto(n).disp = 0
               hdv = dataViewSto(n).hDlg
               hEdt = dataViewSto(n).hEdt
               viewII = dataViewSto(n).viewII
               viewJJ = dataViewSto(n).viewJJ
               row = dataViewSto(n).row
               col = dataViewSto(n).col
               IF wParam& = %VK_RETURN THEN
                  IF dataViewSto(n).nCols < 0 THEN EXIT SELECT
                  GetWindowText hEdt, BYVAL VARPTR(txt), 128
                  s = txt
                  nFont = dataViewSto(n).nFont
                  colWi = dataViewSto(n).colWi
                  getTextSize hdv, x, y, nFont, s
                  getTextSize hdv, x1, y, nFont, SPACE$(1)
                  dataViewEd = s + SPACE$((colWi - 2 - x)/x1)
                  dataViewStr(row, col) = s
                  INCR row
               ELSE
                  s = dataViewStr(row, col)
                  txt = s: SetWindowText hEdt, txt
               END IF
               IF wParam& = %VK_UP THEN
                  IF dataViewSto(n).nCols < 0 THEN
                     DIALOG SEND hdv, %WM_VSCROLL, %SB_LINEUP, 0: EXIT SELECT
                  END IF
                  IF row > 1 THEN DECR row
               ELSEIF wParam& = %VK_DOWN THEN
                  IF dataViewSto(n).nCols < 0 THEN
                     DIALOG SEND hdv, %WM_VSCROLL, %SB_LINEDOWN, 0: EXIT SELECT
                  END IF
                  IF row < UBOUND(dataViewStr, 1) THEN INCR row
               ELSEIF wParam& = %VK_PGUP THEN
                  IF dataViewSto(n).nCols < 0 THEN
                     DIALOG SEND hdv, %WM_HSCROLL, %SB_LINELEFT, 0: EXIT SELECT
                  END IF
                  IF col > 1 THEN DECR col
               ELSEIF wParam& = %VK_PGDN THEN
                  IF dataViewSto(n).nCols < 0 THEN
                     DIALOG SEND hdv, %WM_HSCROLL, %SB_LINERIGHT, 0: EXIT SELECT
                  END IF
                  IF col < UBOUND(dataViewStr, 2) THEN INCR col
               ELSEIF wParam& = %VK_HOME THEN
                  row = 1: col = 1
                  DIALOG SEND hdv, %WM_VSCROLL, %SB_TOP, 0
                  DIALOG SEND hdv, %WM_HSCROLL, %SB_LEFT, 0
                  IF dataViewSto(n).nCols < 0 THEN EXIT SELECT
               END IF
               'negative values for row and col used here
               'to signal dataViewEdit sub not to show text
               IF row > UBOUND(dataViewStr, 1) THEN
                  row = -1
                  DIALOG SEND hdv, %WM_VSCROLL, %SB_TOP, 0
                  INCR col
               ELSEIF row < dataViewSto(n).vPos THEN
                  DIALOG SEND hdv, %WM_VSCROLL, %SB_LINEUP, 0
               ELSEIF row = dataViewSto(n).vPos + viewII THEN
                  DIALOG SEND hdv, %WM_VSCROLL, %SB_LINEDOWN, 0
               END IF
               IF col > UBOUND(dataViewStr, 2) THEN
                  col = -1
                  DIALOG SEND hdv, %WM_HSCROLL, %SB_LEFT, 0
               ELSEIF col < dataViewSto(n).hPos THEN
                  DIALOG SEND hdv, %WM_HSCROLL, %SB_LINELEFT, 0
               ELSEIF col = dataViewSto(n).hPos + viewJJ THEN
                  DIALOG SEND hdv, %WM_HSCROLL, %SB_LINERIGHT, 0
               END IF
               dataViewSto(n).row = row: dataViewSto(n).col = col
               dataViewEdit dataViewSto(n).hEdt
         END SELECT
      CASE %WM_DESTROY
         SetWindowLong hWnd&, %GWL_WNDPROC, GetProp(hWnd&, szOrigWndProc)
         RemoveProp hWnd&, szOrigWndProc: EXIT FUNCTION
   END SELECT
   'pass the message on to the original window procedure... the DDT engine!
   n = GetProp(hWnd&, szOrigWndProc)
   FUNCTION = CallWindowProc(BYVAL n, hWnd&, wMsg&, wParam&, lParam&)
END FUNCTION

FUNCTION MakeFont(hDC AS LONG, sFont AS STRING, BYVAL ptSize AS LONG, _
                  OPT BYVAL descr AS STRING) AS LONG
   'specify descr = b for bold, u for underline, i for italic
   LOCAL CyPixels AS LONG, charHt AS LONG, wt AS DWORD
   LOCAL bold AS DWORD, underline AS DWORD, italic AS DWORD
   IF LEN(descr) THEN
      bold = INSTR(descr, "b") + 300
      underline = INSTR(descr,"u")
      italic = INSTR(descr,"i")
   END IF
   CyPixels  = GetDeviceCaps(hDC, %LOGPIXELSY)
   charHt = (PtSize * CyPixels) \ 72
   wt = %FW_NORMAL + bold
   FUNCTION = CreateFont(-charHt, 0, 0, 0, wt, italic, underline, 0, _
              %ANSI_CHARSET, %OUT_TT_PRECIS, %CLIP_DEFAULT_PRECIS, _
              %DEFAULT_QUALITY, %FF_DONTCARE, BYCOPY sFont)
END FUNCTION
